;;; prop-mode.el --- Parallax Propeller spin and assembler language mode

;; Copyright (C) 2010  Free Software Foundation, Inc.

;; Author: Heiko Schmidt <Heiko.Schmidt@webbedtables.de>
;; Keywords: languages

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; this mode is used for editing Parallax Propeller Spin and
;; assembler programs.

;;; customization
(defgroup prop-mode nil
	"Propeller mode customization group."
	:group 'Languages)

(defcustom prop-mode-std-fontname "Parallax"
	"The name of the prop-mode.el standard font.")

(defcustom prop-mode-std-fontsize 100
	"The standard font size used in prop-mode.el.")

(defcustom prop-mode-compile-binary "bstc"
;;(defcustom prop-mode-compile-binary "propellent.exe"
	"Name of propeller compile tool. On Linux BradC's famous bstc.")

(defcustom prop-mode-serial-interface "/dev/ttyUSB0"
  "Name of the serial interface.")

;;; Code:
(defvar mode-name "Propeller"
	"Pretty printed mode-name. Also indicator for mode is switched
	on or off (nil).")

(defvar prop-mode-syntax-table nil
  "Syntax Table  for Propeller Mode.")

(defvar buffer-face-mode-face nil
	"Dummy definition for later access in prop-mode-face-activate.")

(defvar prop-mode-compile-buffer-name "*Propeller Compile*"
	"prop-mode.el's standard buffer name for compilation output.")

(defun prop-mode-syntax-table-init ()
  "initialize syntax table  for Propeller Mode."
	(let ((table (make-syntax-table)))

	;; Code -Comments with { }
	(modify-syntax-entry ?\{ "< n" table)
	(modify-syntax-entry ?\} "> n" table)
	;; define comment for style : "bla bla ' comment"
	(modify-syntax-entry ?\' "< b" table)
	(modify-syntax-entry ?\n "> b" table)
	table))

(defvar prop-mode-map nil
		"Keymap for prop-mode.
     Many other modes, such as Mail mode, Outline mode
     and Indented Text mode, inherit all the commands
     defined in this map.")

(defun prop-mode-map-init ()
  "Initialize prop-mode-map."
  (setq prop-mode-map (make-sparse-keymap))
  ;; template key bindings
  (define-key prop-mode-map "\C-c\C-s"	   'prop-info-connected-props)
  (define-key prop-mode-map " "           'uppercase-spin-keyword)
  (define-key prop-mode-map "\C-cc"	   'prop-compile)
  (define-key prop-mode-map "\C-cr"	   'prop-compile-and-load-RAM)
  (define-key prop-mode-map "\C-ce"	   'prop-compile-and-load-EEPROM)
	)

(defvar prop-block-designators-keywords
  '("CON"															 ; Constants
    "VAR"															 ; Variables
    "DAT"															 ; Data
    "OBJ"															 ; Objects
    "PUB"															 ; Public
    "PRI"															 ; Privatea
    )
  "*Spin reserved words (will be capitalized)")


(defvar prop-configuration-keywords
	'(
		"CHIPVER"													 ; Propeller chip version number
		"_CLKFREQ" ; Settable in Top Object File to specify System Clock frequency.
		"_CLKMODE" ; Settable in Top Object File to specify application's clock mode.
		"_XINFREQ" ; Settable in Top Object File to specify external crystal frequency.
		"_FREE"	; Settable in Top Object File to specify application's free space.
		"_STACK" ; Settable in Top Object File to specify application's stack space.
		"RCFAST" ; Internal fast oscillator: $00000001 (%00000000001)
		"RCSLOW" ; Internal slow oscillator: $00000002 (%00000000010)
		"XINPUT" ; External clock/oscillator: $00000004 (%00000000100)
		"XTAL1"	 ; External low-speed crystal: $00000008 (%00000001000)
		"XTAL2"	 ; External medium-speed crystal: $00000010 (%00000010000)
		"XTAL3"	 ; External high-speed crystal: $00000020 (%00000100000)
		"PLL1X"	 ; External frequency times 1: $00000040 (%00001000000)
		"PLL2X"	 ; External frequency times 2: $00000080 (%00010000000)
		"PLL4X"	 ; External frequency times 4: $00000100 (%00100000000)
		"PLL8X"	 ; External frequency times 8: $00000200 (%01000000000)
		"PLL16X" ; External frequency times 16: $00000400 (%10000000000)
		)
	"* predefined constants for CON Block")


(defvar prop-types-keywords
	'(
		"long"
		"word"
		"byte"
		)
	"* Spin language types")


(defvar prop-cog-control-keywords
	'(
		"GID"											; Get current cog's ID (0-7).
		"COGNEW"									; Start the next available cog.
		"COGINIT"									; Start, or restart, a cog by ID.
		"COGSTOP"									; Stop a cog by ID.
		"REBOOT"									; Reset the Propeller chip.
		)
	"Spin Cog Control Keywords")

(defvar prop-process-control-keywords
	'(
		"LOCKNEW"							 ; Check out a new semaphore.
		"LOCKRET"							 ; Return a semaphore.
		"LOCKCLR"							 ; Clear a semaphore by ID.
		"LOCKSET"							 ; Set a semaphore by ID.
		"WAITCNT"							 ; Wait for System Counter to reach a value.
		"WAITPEQ"							 ; Wait for pin(s) to be equal to value.
		"WAITPNE"							 ; Wait for pin(s) to be not equal to value.
		"WAITVID" ; Wait for video sync and deliver next color/pixel group.
		)
	"Spin Process Control Keywords")

(defvar prop-flow-control-keywords
	'(
		"IF"					 ; Conditionally execute one or more blocks of code.
		"...ELSEIF"		 ;
		"...ELSE"			 ;
		"CASE" ; Evaluate expression and execute block of code that satisfies a condition.
		"...OTHER"													;
		"REPEAT" ; Execute block of code repetitively an infinite or finite number of times with
		"...FROM" ; optional loop counter, intervals, exit and continue conditions.
		"...TO"		;
		"...STEP" ;
		"...UNTIL"													;
		"...WHILE"													;
		"NEXT" ; Skip rest of REPEAT block and jump to next loop iteration.
		"QUIT" ; Exit from REPEAT loop.
		"RETURN" ; Exit PUB/PRI with normal status and optional return value.
		"ABORT" ; Exit PUB/PRI with abort status and optional return value.
		)
	"Spin Flow Control Keywords")

(defvar prop-memory-keywords
	'(
		"BYTE"	; Declare byte-sized symbol or access byte of main memory.
		"WORD"	; Declare word-sized symbol or access word of main memory.
		"LONG"	; Declare long-sized symbol or access long of main memory.
		"BYTEFILL" ; Fill bytes of main memory with a value.
		"WORDFILL" ; Fill words of main memory with a value.
		"LONGFILL" ; Fill longs of main memory with a value.
		"BYTEMOVE" ; Copy bytes from one region to another in main memory.
		"WORDMOVE" ; Copy words from one region to another in main memory.
		"LONGMOVE" ; Copy longs from one region to another in main memory.
		"LOOKUP"	 ; Get value at index (1-N) from a list.
		"LOOKUPZ"	 ; Get value at zero-based index (0-N) from a list.
		"LOOKDOWN" ; Get index (1-N) of a matching value from a list.
		"LOOKDOWNZ" ; Get zero-based index (0-N) of a matching value from a list.
		"STRSIZE"		; Get size of string in bytes.
		"STRCOMP" ; Compare a string of bytes against another string of bytes.
		)
	"Spin Memory Keywords")

(defvar prop-directives-keywords
	'(
		"STRING" ; Declare in-line string expression; resolved at compile-time.
		"CONSTANT" ; Declare in-line constant expression; resolved at compile-time.
		"FLOAT" ; Declare floating-point expression; resolved at compile-time.
		"ROUND" ; Round compile-time floating point expression to integer.
		"TRUNC" ; Truncate compile-time floating point expression at decimal.
		"FILE"	; Import data from an external file.
		)
	"Spin Directives Keywords")

(defvar prop-registers-keywords
	'(
		"DIRA"				; Direction Register for 32-bit port A.
		"DIRB"				; Direction Register for 32-bit port B (future use).
		"INA"					; Input Register for 32-bit port A (read only).
		"INB"	; Input Register for 32-bit port B (read only) (future use).
		"OUTA"					 ; Output Register for 32-bit port A.
		"OUTB"					 ; Output Register for 32-bit port B (future use).
		"CNT"						 ; 32-bit System Counter Register (read only).
		"CTRA"					 ; Counter A Control Register.
		"CTRB"					 ; Counter B Control Register.
		"FRQA"					 ; Counter A Frequency Register.
		"FRQB"					 ; Counter B Frequency Register.
		"PHSA"					 ; Counter A Phase Lock Loop (PLL) Register.
		"PHSB"					 ; Counter B Phase Lock Loop (PLL) Register.
		"VCFG"					 ; Video Configuration Register.
		"VSCL"					 ; Video Scale Register.
		"PAR"						 ; Cog Boot Parameter Register (read only).
		"SPR" ; Special-Purpose-Register array; indirect cog register access.
		)
	"Spin registers Keywords")

(defvar prop-constants-keywords
	'(
		"TRUE"		 ; Logical True: -1 ($FFFFFFFF).
		"FALSE"		 ; Logical False: 0 ($00000000).
		"POSX"		 ; Maximum positive integer: 2,147,483,647 ($7FFFFFFF).
		"NEGX"		 ; Maximum negative integer: -2,147,483,648 ($80000000).
		"PI"			 ; Floating point value for PI: ~3.141593 ($40490FDB).
		)
	"Spin Constants Keywords")

(defvar prop-variables-keywords
	'(
		"RESULT"						; Default result variable for PUB/PRI methods.
		)
	"Spin Variables Keywords")


;; (defvar prop-font-lock-syntactic-keywords
;; 	'(("^Comments\\(\n\\)" (1 "< b"))
;; 		("EndComments$"
;; 		 (0 (unless (eq (match-beginning 0) (point-min))
;; 					(put-text-property (1- (match-beginning 0)) (match-beginning 0)
;; 														 'syntax-table (eval-when-compile
;; 																						 (string-to-syntax "> b")))
;; 					(put-text-property (1- (match-beginning 0)) (match-end 0)
;; 														 'font-lock-multiline t)
;; 					nil)))))


;; the ','-expression (,(regexp-opt prop-block-designators-keywords
;; 'words) . font-lock-keyword-face) means that it will be evaluated
;; immeadiately at assignment
;; ATTENTION: remind usage of backquote "`" instead of "'" !!!
(defvar prop-font-lock-keywords
	`(
	 (,(regexp-opt prop-block-designators-keywords 'words) .   font-lock-keyword-face)
	 (,(regexp-opt prop-configuration-keywords 'words) .       font-lock-constant-face)
	 (,(regexp-opt prop-types-keywords 'words) .               font-lock-type-face)
	 (,(regexp-opt prop-cog-control-keywords 'words) .         font-lock-keyword-face)
	 (,(regexp-opt prop-process-control-keywords 'words) .     font-lock-keyword-face)
	 (,(regexp-opt prop-flow-control-keywords 'words) .        font-lock-keyword-face)
	 (,(regexp-opt prop-memory-keywords 'words) .              font-lock-keyword-face)
	 (,(regexp-opt prop-directives-keywords 'words) .          font-lock-keyword-face)
	 (,(regexp-opt prop-registers-keywords 'words) .           font-lock-constant-face)
	 (,(regexp-opt prop-constants-keywords 'words) .           font-lock-constant-face)
	 (,(regexp-opt prop-variables-keywords 'words) .           font-lock-variable-name-face)


    ;; note: order above matters. ''mylsl-keywords-regexp'' goes last because
    ;; otherwise the keyword ''state'' in the function ''state_entry''
    ;; would be highlighted.
	 '("\"[^\"]*\"" 0 font-lock-string-face t)
	 '(":=" 0 font-lock-keyword-face t)
	 '("'.*$" 0 font-lock-comment-face t)
;; 		'(font-lock-syntactic-keywords . prop-font-lock-syntactic-keywords)
		))


(defun prop-info-connected-props ()
"Find interface to which a propeller is connected.
NOT YET IMPLEMENTED!"
	(interactive)
	(message "jetzt muessen angeschlossene Propeller gesucht werden")
	)


;; (defun prop-compile ()
;; 	"Compile file (save before compiling!) to <filename>.binary
;; This is used as a syntax check before loading to propeller. bstc
;; output goes to buffer *Propeller Compile*"
;; 	(interactive)
;; 	(let (buffer)
;; 		(erase-prop-compile-buffer)
;; 		(setq  buffer "*Propeller Compile*")
;; 		(call-process prop-mode-compile-binary nil buffer nil "-d" prop-mode-serial-interface "-b" (buffer-file-name) )))

(defun prop-compile ()
	"Compile file (save before compiling!) to <filename>.binary This is
used as a syntax check before loading to propeller. bstc output goes
to buffer prop-mode-compile-buffer-name"
	(interactive)
	(progn
		(erase-prop-compile-buffer)
		(call-process prop-mode-compile-binary nil prop-mode-compile-buffer-name nil "-d" prop-mode-serial-interface "-b" (buffer-file-name) )))

(defun prop-compile-and-load-RAM ()
	"Compile file (save before compiling!) and download to propeller
RAM. bstc output goes to buffer prop-mode-compile-buffer-name"
	(interactive)
	(message "compile and download buffer")
	(progn
		(erase-prop-compile-buffer)
		(call-process prop-mode-compile-binary nil prop-mode-compile-buffer-name nil "-d" prop-mode-serial-interface "-p0" (buffer-file-name) )))

(defun prop-compile-and-load-EEPROM ()
	"Compile file (save before compiling!) and download to propeller
EEPROM. bstc output goes to buffer prop-mode-compile-buffer-name"
	(interactive)
	(progn
		(erase-prop-compile-buffer)
		(call-process prop-mode-compile-binary nil prop-mode-compile-buffer-name nil "-d" prop-mode-serial-interface "-p2" (buffer-file-name) )))

(defun erase-prop-compile-buffer ()
	"Support function for prop-mode compilation functions. Clears buffer
prop-mode-compile-buffer-name (before calling
prop-mode-compile-binary)."

	(progn
		(if (get-buffer prop-mode-compile-buffer-name)
				(save-excursion								 ;save point and mark in current buffer
					(switch-to-buffer prop-mode-compile-buffer-name)						 ; select that buffer
					(erase-buffer)											 ; clear it
					(switch-to-buffer (other-buffer) ; back to edit current edit buffer
														))

			(message "buffer '%s' not existing, so content not erased" prop-mode-compile-buffer-name))))



(defun uppercase-spin-keyword ()
  "check if last word was a spin keyword"
  (interactive)
  (spin-keyword (append prop-block-designators-keywords))
    (insert " "))



(defun spin-keyword (keywords)
  "upcase current word and return 't if it is a spin keyword;
 return nil if it is not"
  (save-excursion
    (let ((eow (point)))
      (forward-word -1)
      (let ((bow (point)))
        (if (re-search-backward "\<" (- bow (current-column)) t)
            nil
          (if (word-in-list (upcase (buffer-substring bow eow))
                            keywords)
              ;;if it is a keyword then
              ;; if we are not in a comment uppercase
              (if (re-search-backward "--" (- (point) (current-column)) t)
                  't
                (not (upcase-region bow eow)))
            nil))))))

(defun word-in-list (word words)
  "t if word occurs in words, nil otherwise"
  (if (null words)
      nil
    (if (string-equal word (car words))
        t
      (word-in-list word (cdr words)))))


(defun prop-mode-face-activate ()
   "Set font to a variable width (proportional) fonts in current buffer"
   (interactive)
;;   (setq buffer-face-mode-face '(:family "Parallax"))
   (setq buffer-face-mode-face `(:family ,prop-mode-std-fontname
																 :height ,prop-mode-std-fontsize))
   (buffer-face-mode))

(defun prop-mode ()
  "Parallax Propeller spin and propeller assemembler language mode
   \\{prop-mode-map}
   Turning on Prop mode runs the normal hook prop-mode-hook."
  (interactive)
	(if (equal major-mode 'prop-mode)
	(progn
		;; prop-mode aus
		(kill-all-local-variables)
		(fundamental-mode)
		)
	(progn
		;; prop-mode ein
		(kill-all-local-variables)
		(setq major-mode 'prop-mode)
;;		(setq mode-name "Propeller")

		;; syntax table
		(setq prop-mode-syntax-table (prop-mode-syntax-table-init))
		(set-syntax-table prop-mode-syntax-table)

		;; set maps and tables
		(prop-mode-map-init)
		(use-local-map prop-mode-map)

		(setq font-lock-defaults '(prop-font-lock-keywords))
		(prop-mode-face-activate)
		(make-local-variable 'prop-mode-serial-interface)
)))

(setq auto-mode-alist (cons '("\\.spin$" . prop-mode) auto-mode-alist))

(provide 'prop-mode)
;;; prop-mode.el ends here
